import numpy as np

from bokeh.models import ColumnDataSource, DataRange1d, Band, Whisker, HoverTool
from bokeh.palettes import Category10_10
from bokeh.plotting import figure

from ..buki import Object
from ..utils import make_tooltips


BLUE = Category10_10[0]
RED  = Category10_10[3]

TOOLTIPS = make_tooltips(
    x = "$x",
    y = "$y",
)


class Plot1D(Object):

    def __init__(self, name=None):
        data = {
            "x": [],
            "y": [],
#            "y-std": [],
#            "y+std": [],
            "avg": [],
#            "avg-err": [],
#            "avg+err": [],
        }

        self.source = source = ColumnDataSource(data=data)

        y_range = DataRange1d(only_visible=True, range_padding=0.5)
        fig = figure(name=name, y_range=y_range)
        fig.toolbar.logo = None

        hover = HoverTool(tooltips=TOOLTIPS, mode="vline")
        fig.add_tools(hover)
        fig.toolbar.active_inspect = None

        self.line_latest,  self.circle_latest,  self.errband_latest,  self.errbars_latest  = add_curve(fig, source, "x", "y", "y-std", "y+std", BLUE)
        self.line_average, self.circle_average, self.errband_average, self.errbars_average = add_curve(fig, source, "x", "avg", "avg-err", "avg+err", RED)

        self.state_errbands = True
        self.state_errbars  = False
        self._set_show_errors()

        super().__init__(fig)


    def set(self, _times, values):
        y = values[-1]

        if y is None: #TODO: is this needed?
            return

        n = len(values)
        values = np.asarray(values)

        x = np.arange(len(y))
        std = values.std(axis=0)

        #TODO: skip avg for n=1 ?
        avg = values.mean(axis=0)
        err = std / np.sqrt(n)

        data = {
            "x": x,
            "y": y,
            "y-std": y - std,
            "y+std": y + std,
            "avg": avg,
            "avg-err": avg - err,
            "avg+err": avg + err,
        }

        self.source.data.update(data)


    def show_curve_latest(self, state):
        self.line_latest.visible = self.circle_latest.visible = state
        self._set_show_errors()

    def show_curve_average(self, state):
        self.line_average.visible = self.circle_average.visible = state
        self._set_show_errors()

    def show_error_bands(self, state):
        self.state_errbands = state
        self.errband_latest.visible  = self.line_latest.visible  and state
        self.errband_average.visible = self.line_average.visible and state

    def show_error_bars(self, state):
        self.state_errbars = state
        self.errbars_latest.visible  = self.line_latest.visible  and state
        self.errbars_average.visible = self.line_average.visible and state

    def _set_show_errors(self):
        self.show_error_bands(self.state_errbands)
        self.show_error_bars(self.state_errbars)



def add_curve(fig, source, x, y, lower, upper, color):
    line = fig.line(x=x, y=y, source=source, color=color)
    circle = fig.circle(x=x, y=y, source=source, color=color)

    errband = Band(base=x, lower=lower, upper=upper, source=source, level="underlay", line_width=1, line_color=color, fill_color=color, fill_alpha=0.5)

    errbars = Whisker(base=x, lower=lower, upper=upper, source=source, line_color=color)
    errbars.lower_head.line_color = color
    errbars.upper_head.line_color = color

    fig.add_layout(errband)
    fig.add_layout(errbars)

    return line, circle, errband, errbars



